from datetime import datetime
from typing import Optional

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from edc_appointment.constants import COMPLETE_APPT, IN_PROGRESS_APPT
from edc_appointment.creators import AppointmentsCreator
from edc_consent import NotConsentedError
from edc_utils import convert_php_dateformat, formatted_datetime, get_utcnow

from .constants import OFF_SCHEDULE, ON_SCHEDULE
from .exceptions import (
    InvalidOffscheduleDate,
    NotOnScheduleError,
    NotOnScheduleForDateError,
    OnScheduleFirstAppointmentDateError,
    SubjectScheduleError,
    UnknownSubjectError,
)


class SubjectSchedule:
    """A class that puts a subject on to a schedule or takes a subject
    off of a schedule.

    This class is instantiated by the Schedule class.
    """

    history_model = "edc_visit_schedule.subjectschedulehistory"
    registered_subject_model = "edc_registration.registeredsubject"
    appointments_creator_cls = AppointmentsCreator

    def __init__(self, visit_schedule=None, schedule=None):
        self.visit_schedule = visit_schedule
        self.schedule = schedule
        self.schedule_name = schedule.name
        self.visit_schedule_name = self.visit_schedule.name
        self.onschedule_model = schedule.onschedule_model
        self.consent_model = schedule.consent_model
        self.offschedule_model = schedule.offschedule_model
        self.appointment_model = schedule.appointment_model

    @property
    def onschedule_model_cls(self):
        return django_apps.get_model(self.onschedule_model)

    @property
    def offschedule_model_cls(self):
        return django_apps.get_model(self.offschedule_model)

    @property
    def history_model_cls(self):
        return django_apps.get_model(self.history_model)

    @property
    def appointment_model_cls(self):
        return django_apps.get_model(self.appointment_model)

    @property
    def visit_model_cls(self):
        return self.appointment_model_cls.visit_model_cls()

    @property
    def consent_model_cls(self):
        return django_apps.get_model(self.consent_model)

    def put_on_schedule(
        self,
        subject_identifier: Optional[str] = None,
        onschedule_datetime: Optional[datetime] = None,
        first_appt_datetime: Optional[datetime] = None,
        skip_baseline: Optional[bool] = None,
    ):
        """Puts a subject on-schedule.

        A person is put on schedule by creating an instance
        of the onschedule_model, if it does not already exist,
        and updating the history_obj.
        """
        onschedule_datetime = onschedule_datetime or get_utcnow()
        if not self.onschedule_model_cls.objects.filter(
            subject_identifier=subject_identifier
        ).exists():
            self.registered_or_raise(subject_identifier=subject_identifier)
            self.consented_or_raise(subject_identifier=subject_identifier)
            self.onschedule_model_cls.objects.create(
                subject_identifier=subject_identifier,
                onschedule_datetime=onschedule_datetime,
            )
        try:
            history_obj = self.history_model_cls.objects.get(
                subject_identifier=subject_identifier,
                schedule_name=self.schedule_name,
                visit_schedule_name=self.visit_schedule_name,
            )
        except ObjectDoesNotExist:
            history_obj = self.history_model_cls.objects.create(
                subject_identifier=subject_identifier,
                onschedule_model=self.onschedule_model,
                offschedule_model=self.offschedule_model,
                schedule_name=self.schedule_name,
                visit_schedule_name=self.visit_schedule_name,
                onschedule_datetime=onschedule_datetime,
                schedule_status=ON_SCHEDULE,
            )
        if history_obj.schedule_status == ON_SCHEDULE:
            # create appointments per schedule
            creator = self.appointments_creator_cls(
                report_datetime=onschedule_datetime,
                subject_identifier=subject_identifier,
                schedule=self.schedule,
                visit_schedule=self.visit_schedule,
                appointment_model=self.appointment_model,
                skip_baseline=skip_baseline,
            )
            if first_appt_datetime and first_appt_datetime < onschedule_datetime:
                raise OnScheduleFirstAppointmentDateError(
                    "First appt datetime cannot be before onschedule datetime. "
                    f"Got {first_appt_datetime} < {onschedule_datetime}"
                )
            creator.create_appointments(first_appt_datetime or onschedule_datetime)

    def take_off_schedule(self, subject_identifier=None, offschedule_datetime=None):
        """Takes a subject off-schedule.

        A person is taken off-schedule by:
        * creating an instance of the offschedule_model,
          if it does not already exist,
        * updating the history_obj
        * deleting future appointments
        """
        # create offschedule_model_obj if it does not exist
        if not self.offschedule_model_cls.objects.filter(
            subject_identifier=subject_identifier
        ).exists():
            self.offschedule_model_cls.objects.create(
                subject_identifier=subject_identifier,
                offschedule_datetime=offschedule_datetime,
            )

        # get existing history obj or raise
        try:
            history_obj = self.history_model_cls.objects.get(
                subject_identifier=subject_identifier,
                schedule_name=self.schedule_name,
                visit_schedule_name=self.visit_schedule_name,
            )
        except ObjectDoesNotExist:
            raise NotOnScheduleError(
                "Failed to take subject off schedule. "
                f"Subject has not been put on schedule "
                f"'{self.visit_schedule_name}.{self.schedule_name}'. "
                f"Got '{subject_identifier}'."
            )

        if history_obj:
            self.update_history_or_raise(
                history_obj=history_obj,
                subject_identifier=subject_identifier,
                offschedule_datetime=offschedule_datetime,
            )

            self._update_in_progress_appointment(subject_identifier=subject_identifier)

            # clear future appointments
            self.appointment_model_cls.objects.delete_for_subject_after_date(
                subject_identifier=subject_identifier,
                cutoff_datetime=offschedule_datetime,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
            )

    def update_history_or_raise(
        self,
        history_obj=None,
        subject_identifier=None,
        offschedule_datetime=None,
        update=None,
    ):
        """Updates the history model instance.

        Raises an error if the offschedule_datetime is before the
        onschedule_datetime or before the last visit.
        """
        update = True if update is None else update
        if not self.history_model_cls.objects.filter(
            subject_identifier=subject_identifier,
            schedule_name=self.schedule_name,
            visit_schedule_name=self.visit_schedule_name,
            onschedule_datetime__lte=offschedule_datetime,
        ).exists():
            raise InvalidOffscheduleDate(
                "Failed to take subject off schedule. "
                "Offschedule date cannot precede onschedule date. "
                f"Subject was put on schedule {self.visit_schedule_name}."
                f"{self.schedule_name} on {history_obj.onschedule_datetime}. "
                f"Got {offschedule_datetime}."
            )
        # confirm date not before last visit
        related_visit_model_attr = self.appointment_model_cls.related_visit_model_attr()
        try:
            appointments = self.appointment_model_cls.objects.get(
                subject_identifier=subject_identifier,
                schedule_name=self.schedule_name,
                visit_schedule_name=self.visit_schedule_name,
                **{f"{related_visit_model_attr}__report_datetime__gt": offschedule_datetime},
            )
        except ObjectDoesNotExist:
            appointments = None
        except MultipleObjectsReturned:
            appointments = self.appointment_model_cls.objects.filter(
                subject_identifier=subject_identifier,
                schedule_name=self.schedule_name,
                visit_schedule_name=self.visit_schedule_name,
                **{f"{related_visit_model_attr}__report_datetime__gt": offschedule_datetime},
            )
        if appointments:
            raise InvalidOffscheduleDate(
                f"Failed to take subject off schedule. "
                f"Visits exist after proposed offschedule date. "
                f"Got '{formatted_datetime(offschedule_datetime)}'."
            )
        if update:
            # update history object
            history_obj.offschedule_datetime = offschedule_datetime
            history_obj.schedule_status = OFF_SCHEDULE
            history_obj.save()

    def _update_in_progress_appointment(self, subject_identifier=None):
        """Updates the "in_progress" appointment and clears
        future appointments.
        """
        for obj in self.appointment_model_cls.objects.filter(
            subject_identifier=subject_identifier,
            schedule_name=self.schedule_name,
            visit_schedule_name=self.visit_schedule_name,
            appt_status=IN_PROGRESS_APPT,
        ):
            obj.appt_status = COMPLETE_APPT
            obj.save()

    def resave(self, subject_identifier=None):
        """Resaves the onschedule model instance to trigger, for example,
        appointment creation (if using edc_appointment mixin).
        """
        obj = self.onschedule_model_cls.objects.get(subject_identifier=subject_identifier)
        obj.save()

    def registered_or_raise(self, subject_identifier=None):
        """Raises an exception if RegisteredSubject instance does not exist."""
        model_cls = django_apps.get_model(self.registered_subject_model)
        try:
            model_cls.objects.get(subject_identifier=subject_identifier)
        except ObjectDoesNotExist:
            raise UnknownSubjectError(
                f"Failed to put subject on schedule. Unknown subject. "
                f"Searched `{self.registered_subject_model}`. "
                f"Got subject_identifier=`{subject_identifier}`."
            )

    def consented_or_raise(self, subject_identifier=None):
        """Raises an exception if one or more consents do not exist."""
        if not self.consent_model_cls.objects.filter(
            subject_identifier=subject_identifier
        ).exists():
            raise NotConsentedError(
                f"Failed to put subject on schedule. Consent not found. "
                f"Using consent model '{self.consent_model}' "
                f"subject identifier={subject_identifier}."
            )

    def onschedule_or_raise(
        self, subject_identifier=None, report_datetime=None, compare_as_datetimes=None
    ):
        """Raise an exception if subject is not on the schedule during
        the given date.
        """
        compare_as_datetimes = True if compare_as_datetimes is None else compare_as_datetimes
        try:
            onschedule_obj = self.onschedule_model_cls.objects.get(
                subject_identifier=subject_identifier
            )
        except ObjectDoesNotExist:
            raise NotOnScheduleError(
                f"Subject has not been put on a schedule `{self.schedule_name}`. "
                f"Got subject_identifier=`{subject_identifier}`."
            )

        try:
            offschedule_datetime = self.offschedule_model_cls.objects.values_list(
                "offschedule_datetime", flat=True
            ).get(subject_identifier=subject_identifier)
        except ObjectDoesNotExist:
            offschedule_datetime = None

        if compare_as_datetimes:
            in_date_range = (
                onschedule_obj.onschedule_datetime
                <= report_datetime
                <= (offschedule_datetime or get_utcnow())
            )
        else:
            in_date_range = (
                onschedule_obj.onschedule_datetime.date()
                <= report_datetime.date()
                <= (offschedule_datetime or get_utcnow()).date()
            )

        if offschedule_datetime and not in_date_range:
            formatted_offschedule_datetime = offschedule_datetime.strftime(
                convert_php_dateformat(settings.SHORT_DATE_FORMAT)
            )
            formatted_report_datetime = report_datetime.strftime(
                convert_php_dateformat(settings.SHORT_DATE_FORMAT)
            )
            raise NotOnScheduleForDateError(
                f"Subject not on schedule '{self.schedule_name}' for "
                f"report date '{formatted_report_datetime}'. "
                f"Got '{subject_identifier}' was taken "
                f"off this schedule on '{formatted_offschedule_datetime}'."
            )
        return None

    def check(self):
        try:
            self.onschedule_model_cls
            self.offschedule_model_cls
            self.appointment_model_cls
        except LookupError as e:
            raise SubjectScheduleError(e)



from django.db import models
from edc_model import models as edc_models

from ..model_mixins import VisitScheduleMethodsModelMixin


class VisitScheduleManager(models.Manager):
    def get_by_natural_key(self, visit_schedule_name, schedule_name, visit_code):
        return self.get(
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
            visit_code=visit_code,
        )


class VisitSchedule(VisitScheduleMethodsModelMixin, edc_models.BaseUuidModel):
    visit_schedule_name = models.CharField(max_length=150)

    schedule_name = models.CharField(max_length=150)

    visit_code = models.CharField(max_length=150)

    visit_name = models.CharField(max_length=150)

    visit_title = models.CharField(max_length=150)

    timepoint = models.DecimalField(null=True, decimal_places=1, max_digits=6)

    active = models.BooleanField(default=False)

    objects = VisitScheduleManager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return f"{self.visit_code}.0: {self.visit_title}"

    def natural_key(self):
        return (
            self.visit_schedule_name,
            self.schedule_name,
            self.visit_code,
        )

    class Meta(edc_models.BaseUuidModel.Meta):
        ordering = ("visit_schedule_name", "schedule_name", "visit_code")
        unique_together = (
            ("visit_schedule_name", "schedule_name", "visit_code"),
            ("visit_schedule_name", "schedule_name", "timepoint"),
        )
        indexes = [
            models.Index(
                fields=[
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_name",
                    "visit_title",
                ]
            ),
            models.Index(fields=["visit_schedule_name", "schedule_name", "timepoint"]),
        ]



from dateutil.relativedelta import relativedelta
from django.apps import apps as django_apps
from django.test import TestCase

from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.system_checks import visit_schedule_check
from edc_visit_schedule.visit import FormsCollection, Visit
from edc_visit_schedule.visit.crf import Crf
from edc_visit_schedule.visit_schedule import VisitSchedule


class TestSystemChecks(TestCase):
    def test_system_check(self):
        site_visit_schedules._registry = {}
        errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "edc_visit_schedule.001")

    def test_visit_schedule_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = VisitSchedule(
            name="visit_schedule",
            verbose_name="Visit Schedule",
            offstudy_model="visit_schedule_app.subjectoffstudy",
            death_report_model="visit_schedule_app.deathreport",
            locator_model="edc_locator.subjectlocator",
        )
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            appointment_model="edc_appointment.appointment",
            consent_model="visit_schedule_app.subjectconsent",
            base_timepoint=1,
        )
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)
        errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(errors), 0)

    def test_visit_schedule_bad_model(self):
        site_visit_schedules._registry = {}
        visit_schedule = VisitSchedule(
            name="visit_schedule",
            verbose_name="Visit Schedule",
            offstudy_model="blah.subjectoffstudy",
            death_report_model="visit_schedule_app.deathreport",
            locator_model="edc_locator.subjectlocator",
        )
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            appointment_model="edc_appointment.appointment",
            consent_model="visit_schedule_app.subjectconsent",
            base_timepoint=1,
        )
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)
        errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(errors), 1)
        self.assertEqual("edc_visit_schedule.visit_schedules", errors[0].id)

    def test_schedule_bad_model(self):
        site_visit_schedules._registry = {}
        visit_schedule = VisitSchedule(
            name="visit_schedule",
            verbose_name="Visit Schedule",
            offstudy_model="visit_schedule_app.subjectoffstudy",
            death_report_model="visit_schedule_app.deathreport",
            locator_model="edc_locator.subjectlocator",
        )
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            appointment_model="blah.appointment",
            consent_model="visit_schedule_app.subjectconsent",
            base_timepoint=1,
        )
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)
        errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(errors), 1)
        self.assertEqual("edc_visit_schedule.schedules", errors[0].id)

    def test_schedule_bad_crf_model(self):
        site_visit_schedules._registry = {}
        visit_schedule = VisitSchedule(
            name="visit_schedule",
            verbose_name="Visit Schedule",
            offstudy_model="visit_schedule_app.subjectoffstudy",
            death_report_model="visit_schedule_app.deathreport",
            locator_model="edc_locator.subjectlocator",
        )
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            appointment_model="edc_appointment.appointment",
            consent_model="visit_schedule_app.subjectconsent",
            base_timepoint=1,
        )
        crfs = FormsCollection(
            Crf(show_order=10, model="blah.CrfOne"),
            Crf(show_order=20, model="blah.CrfTwo"),
            Crf(show_order=30, model="blah.CrfThree"),
        )
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)
        errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(errors), 1)
        self.assertEqual("edc_visit_schedule.visits", errors[0].id)



from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings
from edc_consent import site_consents
from edc_consent.consent import Consent
from edc_constants.constants import FEMALE, MALE
from edc_facility.import_holidays import import_holidays
from edc_protocol import Protocol
from edc_reference import site_reference_configs
from edc_sites.tests import SiteTestCaseMixin
from edc_utils import get_utcnow

from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from visit_schedule_app.forms import OffScheduleForm
from visit_schedule_app.models import OnSchedule, SubjectConsent
from visit_schedule_app.visit_schedule import visit_schedule


@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
)
class TestModels(SiteTestCaseMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        import_holidays()
        super().setUpClass()

    def setUp(self):
        site_visit_schedules.loaded = False
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_visit_tracking.subjectvisit"}
        )
        v1_consent = Consent(
            "visit_schedule_app.subjectconsent",
            version="1",
            start=Protocol().study_open_datetime,
            end=Protocol().study_close_datetime,
            age_min=18,
            age_is_adult=18,
            age_max=64,
            gender=[MALE, FEMALE],
        )
        self.subject_identifier = "1234"
        site_consents.registry = {}
        site_consents.register(v1_consent)

    def test_offschedule_ok(self):
        SubjectConsent.objects.create(subject_identifier=self.subject_identifier)
        onschedule = OnSchedule.objects.create(
            subject_identifier=self.subject_identifier,
        )
        data = dict(
            subject_identifier=self.subject_identifier,
            offschedule_datetime=onschedule.onschedule_datetime + relativedelta(months=1),
        )
        form = OffScheduleForm(data=data)
        form.is_valid()



from __future__ import annotations

import re
from decimal import Decimal
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from edc_facility.utils import get_default_facility_name, get_facility
from edc_utils import get_utcnow, to_utc

from .window_period import WindowPeriod

if TYPE_CHECKING:
    from datetime import datetime

    from dateutil.relativedelta import relativedelta
    from edc_facility import Facility

    from .forms_collection import FormsCollection


class VisitCodeError(Exception):
    pass


class VisitDateError(Exception):
    pass


class VisitError(Exception):
    pass


class BaseDatetimeNotSet(Exception):
    pass


def base_datetime_required():
    raise BaseDatetimeNotSet(
        "Base datetime is None, set the base datetime before accessing attr"
    )


class VisitDate:
    window_period_cls = WindowPeriod

    def __init__(
        self,
        rlower: relativedelta = None,
        rupper: relativedelta = None,
        timepoint: Decimal = None,
        base_timepoint: Decimal = None,
    ):
        self._base: datetime | None = None
        self._lower: datetime | None = None
        self._upper: datetime | None = None
        self._window_period = self.window_period_cls(
            rlower=rlower,
            rupper=rupper,
            timepoint=timepoint,
            base_timepoint=base_timepoint,
        )

    @property
    def base(self) -> datetime:
        return self._base

    @base.setter
    def base(self, dt: datetime = None):
        self._base = to_utc(dt)
        self._lower, self._upper = self._window_period.get_window(dt=self._base)

    @property
    def lower(self) -> datetime:
        if not self.base:
            raise BaseDatetimeNotSet(
                "Base datetime is None, set the base datetime before accessing attr lower"
            )
        return self._lower

    @property
    def upper(self) -> datetime:
        if not self.base:
            raise BaseDatetimeNotSet(
                "Base datetime is None, set the base datetime before accessing attr upper"
            )
        return self._upper


class Visit:
    code_regex = r"^([A-Z0-9])+$"
    visit_date_cls = VisitDate

    def __init__(
        self,
        code: str = None,
        timepoint: int | float | Decimal = None,
        rbase: relativedelta = None,
        rlower: relativedelta = None,
        rupper: relativedelta = None,
        rlower_late: relativedelta = None,
        rupper_late: relativedelta = None,
        add_window_gap_to_lower: bool | None = None,
        max_window_gap_to_lower: int | None = None,
        crfs: FormsCollection | None = None,
        requisitions: FormsCollection | None = None,
        crfs_unscheduled: FormsCollection | None = None,
        crfs_missed: FormsCollection | None = None,
        requisitions_unscheduled: FormsCollection | None = None,
        crfs_prn: FormsCollection | None = None,
        requisitions_prn: FormsCollection | None = None,
        title: str = None,
        allow_unscheduled: bool | None = None,
        facility_name: str | None = None,
        instructions: str | None = None,
        base_timepoint: int | float | Decimal | None = None,
        grouping=None,
    ):
        self.next = None
        if isinstance(base_timepoint, (float,)):
            base_timepoint = Decimal(str(base_timepoint))
        elif isinstance(base_timepoint, (int,)):
            base_timepoint = Decimal(str(base_timepoint) + ".0")
        self.base_timepoint = base_timepoint or Decimal("0.0")
        self.crfs = crfs.forms if crfs else tuple()
        self.crfs_unscheduled = crfs_unscheduled.forms if crfs_unscheduled else ()
        self.crfs_missed = crfs_missed.forms if crfs_missed else ()
        self.crfs_prn = crfs_prn.forms if crfs_prn else ()
        for prn in self.crfs_prn:
            prn.required = False
        self.requisitions = requisitions.forms if requisitions else ()
        self.requisitions_unscheduled = (
            requisitions_unscheduled.forms if requisitions_unscheduled else ()
        )
        self.requisitions_prn = requisitions_prn.forms if requisitions_prn else ()
        for prn in self.requisitions_prn:
            prn.required = False
        self.instructions = instructions
        if isinstance(timepoint, (float,)):
            timepoint = Decimal(str(timepoint))
        elif isinstance(timepoint, (int,)):
            timepoint = Decimal(str(timepoint) + ".0")
        self.timepoint = timepoint
        self.rbase = rbase
        self.rlower = rlower
        self.rupper = rupper
        self.rlower_late = self.rlower if rlower_late is None else rlower_late
        self.rupper_late = self.rupper if rupper_late is None else rupper_late
        self.add_window_gap_to_lower = add_window_gap_to_lower
        self.max_window_gap_to_lower = max_window_gap_to_lower
        self.grouping = grouping
        if not code or isinstance(code, int) or not re.match(self.code_regex, code):
            raise VisitCodeError(f"Invalid visit code. Got '{code}'")
        else:
            self.code = code  # unique
        self.dates = self.visit_date_cls(
            rlower=rlower,
            rupper=rupper,
            timepoint=self.timepoint,
            base_timepoint=self.base_timepoint,
        )
        self.late_dates = self.visit_date_cls(
            rlower=rlower_late or rlower,
            rupper=rupper_late or rupper,
            timepoint=self.timepoint,
            base_timepoint=self.base_timepoint,
        )
        utcnow = get_utcnow()
        self.validate_rlower_late(utcnow)
        self.validate_rupper_late(utcnow)
        self.title = title or f"Visit {self.code}"
        self.name = self.code
        self.facility_name = facility_name or get_default_facility_name()
        self.allow_unscheduled = allow_unscheduled
        if timepoint is None:
            raise VisitError(f"Timepoint not specified. Got None. See Visit {code}.")

    def __repr__(self):
        return f"{self.__class__.__name__}({self.code}, {self.timepoint})"

    def __str__(self):
        return self.title

    @property
    def forms(self):
        """Returns a list of scheduled forms."""
        return self.crfs + self.requisitions

    @property
    def unscheduled_forms(self):
        """Returns a list of unscheduled forms."""
        return self.crfs_unscheduled + self.requisitions_unscheduled

    @property
    def missed_forms(self):
        """Returns a list of forms to show for a missed visit."""
        return self.crfs_missed

    @property
    def prn_forms(self):
        """Returns a list of PRN forms."""
        return self.crfs_prn + self.requisitions_prn

    @property
    def all_crfs(self):
        return self.crfs + self.crfs_unscheduled + self.crfs_prn + self.crfs_missed

    @property
    def all_requisitions(self):
        names = list(set([r.name for r in self.requisitions]))
        requisitions = list(self.requisitions) + [
            r for r in self.requisitions_unscheduled if r.name not in names
        ]
        names = list(set([r.name for r in requisitions]))
        requisitions = requisitions + [r for r in self.requisitions_prn if r.name not in names]
        return sorted(requisitions, key=lambda x: x.show_order)

    def next_form(self, model=None, panel=None):
        """Returns the next required "form" or None."""
        next_form = None
        for index, form in enumerate(self.forms):
            if form.model == model and form.required:
                try:
                    next_form = self.forms[index + 1]
                except IndexError:
                    pass
        return next_form

    def get_form(self, model=None):
        for form in self.forms:
            if form.model == model:
                return form
        return None

    def get_crf(self, model=None):
        get_crf = None
        for crf in self.crfs:
            if crf.model == model:
                get_crf = crf
                break
        return get_crf

    def get_requisition(self, model=None, panel_name=None):
        get_requisition = None
        for requisition in self.requisitions:
            if requisition.model == model and requisition.panel.name == panel_name:
                get_requisition = requisition
                break
        return get_requisition

    @property
    def facility(self) -> Facility | None:
        """Returns a Facility object."""
        if self.facility_name:
            return get_facility(name=self.facility_name)
        return None

    @property
    def timepoint_datetime(self) -> datetime:
        return self.dates.base

    @timepoint_datetime.setter
    def timepoint_datetime(self, dt=None):
        self.dates.base = to_utc(dt)

    def check(self):
        warnings = []
        crf = None
        try:
            for crf in self.crfs:
                django_apps.get_model(crf.model)
            for crf in self.requisitions:
                django_apps.get_model(crf.model)
            for crf in self.crfs_unscheduled:
                django_apps.get_model(crf.model)
            for crf in self.crfs_missed:
                django_apps.get_model(crf.model)
            for crf in self.requisitions_unscheduled:
                django_apps.get_model(crf.model)
        except LookupError as e:
            warnings.append(f"{e} Got Visit {self.code} crf.model={crf.model}.")
        return warnings

    def to_dict(self):
        return dict(
            crfs=[(crf.model, crf.required) for crf in self.crfs],
            crfs_unscheduled=[(crf.model, crf.required) for crf in self.crfs_unscheduled],
            requisitions=[
                (requisition.panel.name, requisition.required)
                for requisition in self.requisitions
            ],
            requisitions_unscheduled=[
                (requisition.panel.name, requisition.required)
                for requisition in self.requisitions_unscheduled
            ],
        )

    def validate_rlower_late(self, utcnow):
        if utcnow - self.rlower_late < utcnow - self.rlower:
            raise VisitDateError(
                "Lower bound error. `Late` lower boundary cannot exceed "
                f"lower boundary. See {self}."
            )

    def validate_rupper_late(self, utcnow):
        if utcnow + self.rupper_late < utcnow + self.rupper:
            raise VisitDateError(
                "Upper bound error. `Late` upper boundary cannot be exceed "
                f"upper boundary. See {self}."
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Self

from django.db import models

from ...site_visit_schedules import site_visit_schedules

if TYPE_CHECKING:
    from edc_visit_schedule.typing_stubs import VisitScheduleFieldsProtocol

    from ...schedule import Schedule, VisitCollection
    from ...visit import Visit
    from ...visit_schedule import VisitSchedule


class VisitScheduleModelMixinError(Exception):
    pass


class VisitScheduleMethodsModelMixin(models.Model):
    """A model mixin that adds methods used to work with the visit schedule.

    Declare with VisitScheduleFieldsModelMixin or the fields from
    VisitScheduleFieldsModelMixin
    """

    @property
    def visit(self) -> Visit:
        """Returns the visit object from the schedule object
        for this visit code.

        Note: This is not a model instance.
        """
        return self.visit_from_schedule

    @property
    def visit_from_schedule(self: VisitScheduleFieldsProtocol | Self) -> Visit:
        """Returns the visit object from the schedule object
        for this visit code.

        Note: This is not a model instance.
        """
        visit = self.schedule.visits.get(self.visit_code)
        if not visit:
            raise VisitScheduleModelMixinError(
                f"Visit not found in schedule. Expected one of {self.schedule.visits}. "
                f"Got {self.visit_code}."
            )
        return visit

    @property
    def visits(self) -> VisitCollection:
        """Returns all visit objects from the schedule object."""
        return self.schedule.visits

    @property
    def schedule(self: VisitScheduleFieldsProtocol | Self) -> Schedule:
        """Returns a schedule object from Meta.visit_schedule_name or
        self.schedule_name.

        Declared on Meta like this:
            visit_schedule_name = 'visit_schedule_name.schedule_name'
        """
        return self.visit_schedule.schedules.get(self.schedule_name)

    @property
    def visit_schedule(self: VisitScheduleFieldsProtocol | Self) -> VisitSchedule:
        """Returns a visit schedule object"""
        return site_visit_schedules.get_visit_schedule(
            visit_schedule_name=self.visit_schedule_name
        )

    class Meta:
        abstract = True



from datetime import date

from django.db import models
from django.db.models.deletion import PROTECT
from edc_crf.model_mixins import CrfModelMixin, CrfWithActionModelMixin
from edc_identifier.managers import SubjectIdentifierManager
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_list_data.model_mixins import ListModelMixin
from edc_model.models import BaseUuidModel
from edc_offstudy.model_mixins import OffstudyModelMixin
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_sites.models import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_tracking.model_mixins import SubjectVisitMissedModelMixin
from edc_visit_tracking.models import SubjectVisit

from edc_visit_schedule.model_mixins import OffScheduleModelMixin, OnScheduleModelMixin


class SubjectVisitMissedReasons(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Subject Missed Visit Reasons"
        verbose_name_plural = "Subject Missed Visit Reasons"


class SubjectVisitMissed(
    SubjectVisitMissedModelMixin,
    CrfWithActionModelMixin,
    BaseUuidModel,
):
    missed_reasons = models.ManyToManyField(
        SubjectVisitMissedReasons, blank=True, related_name="+"
    )

    class Meta(
        SubjectVisitMissedModelMixin.Meta,
        BaseUuidModel.Meta,
    ):
        verbose_name = "Missed Visit Report"
        verbose_name_plural = "Missed Visit Report"


class SubjectScreening(
    SiteModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    BaseUuidModel,
):
    screening_identifier = models.CharField(max_length=50)

    screening_datetime = models.DateTimeField(default=get_utcnow)

    age_in_years = models.IntegerField(default=25)


class SubjectConsent(
    SiteModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    BaseUuidModel,
):
    consent_datetime = models.DateTimeField(default=get_utcnow)

    version = models.CharField(max_length=25, default="1")

    identity = models.CharField(max_length=25)

    confirm_identity = models.CharField(max_length=25)

    dob = models.DateField(default=date(1995, 1, 1))


class SubjectOffstudy(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudyFive(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudySix(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class SubjectOffstudySeven(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    objects = SubjectIdentifierManager()


class DeathReport(SiteModelMixin, BaseUuidModel):
    subject_identifier = models.CharField(max_length=25, null=True)

    report_datetime = models.DateTimeField()


# visit_schedule


class OnSchedule(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffSchedule(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    class Meta(OffScheduleModelMixin.Meta):
        pass


class OnScheduleThree(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleThree(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    class Meta(OffScheduleModelMixin.Meta):
        pass


# visit_schedule_two


class OnScheduleTwo(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleTwo(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    pass


class OnScheduleFour(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleFour(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    pass


class OnScheduleFive(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleFive(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    offschedule_datetime_field_attr = "my_offschedule_datetime"

    my_offschedule_datetime = models.DateTimeField()

    class Meta(OffScheduleModelMixin.Meta):
        pass


class OnScheduleSix(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleSix(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    offschedule_datetime_field_attr = "my_offschedule_date"

    my_offschedule_date = models.DateField()

    class Meta(OffScheduleModelMixin.Meta):
        pass


class BadOffSchedule1(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    """Meta.OffScheduleModelMixin.offschedule_datetime_field
    is None"""

    offschedule_datetime_field_attr = None

    my_offschedule_date = models.DateField()

    class Meta(OffScheduleModelMixin.Meta):
        pass


class OnScheduleSeven(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleSeven(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    """Is Missing Meta.OffScheduleModelMixin"""

    class Meta:
        pass


class CrfOne(CrfModelMixin, BaseUuidModel):
    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)



from django.views.generic.base import TemplateView
from edc_dashboard.utils import get_bootstrap_version
from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar.view_mixin import NavbarViewMixin

from ..site_visit_schedules import SiteVisitScheduleError, site_visit_schedules


class VisitScheduleView(EdcViewMixin, NavbarViewMixin, TemplateView):
    template_name = (
        f"edc_visit_schedule/bootstrap{get_bootstrap_version()}/visit_schedule.html"
    )
    navbar_name = "edc_visit_schedule"
    navbar_selected_item = "visit_schedule"

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        try:
            selected_visit_schedule = site_visit_schedules.get_visit_schedule(
                visit_schedule_name=self.kwargs.get("visit_schedule")
            )
        except SiteVisitScheduleError:
            selected_visit_schedule = None
        context_data.update(
            {
                "visit_schedules": site_visit_schedules.registry,
                "selected_visit_schedule": selected_visit_schedule,
            }
        )
        return context_data



from django import forms

from .models import SubjectScheduleHistory


class SubjectScheduleHistoryForm(forms.ModelForm):
    def clean(self):
        raise forms.ValidationError(
            "This is not a user form. This form may only be edited by the system."
        )

    class Meta:
        model = SubjectScheduleHistory
        fields = "__all__"
